'''
Script of all helper + core data processor class DataProcessor.

Used exclusively data preprocessing (i.e. data.json -> binned jpgs in data/)
'''

import json
import os
import shutil
import numpy as np
import pandas as pd
from PIL import Image

class jsonParser:
    '''
    Parser class to handle relevant data from `data.json` easily.
    
    params:
        - json_path (str): Path to JSON of saved simulation.
    '''
    def __init__(self, json_path:str) -> None:
        self.json_path = json_path

        with open(self.json_path, 'r') as j:
            j = json.load(j)
        
        # Parse settings and data as first level access
        self.settings = j['settings']
        self.data = j['data']

        # Can add more if needed later on.
        self.temp_min = self.settings['tempMin']
        self.temp_max = self.settings['tempMax']
        self.temp_step = self.settings['tempStep']

class nBin:
    '''
    Handles all binning actions including interval creation, data binning, and data alignment.
    
    params:
        - min_ (int/float): Minimum temperature value. Supports both integer and float.
        - max_ (int/float): Maximum temperature value. Supports both integer and float.
        - n_bins (int): Number of bins to split between minimum and maximum.
        - bin_len (float): Length of each bin.
    '''
    def __init__(self, min_, max_, n_bins:int, bin_len:float=None) -> None:
        '''
        params:
            - min_ (int, float): minimum temperature
            - max_ (int, float): maximum temperature
            - n_bins (int): n_bins
            - bin_len (float): float
            '''
        self.min_ = min_
        self.max_ = max_
        self.n_bins = n_bins
        self.bin_len = bin_len
    
    def _calculate_intervals(self):
        '''Calculate bounds of intervals as an array of shape `(n_bins, 2)`, where the columns represent the lower and upper bound respectively.'''
        self.intervals = np.array([
            [self.min_ + (i * self.bin_len) for i in range(self.n_bins)], # Lower bound
            [self.min_ + ((i + 1) * (self.bin_len)) for i in range(self.n_bins)] # Upper bound
        ]).T

class DataProcessor:
    '''
    Handling class to process saved simulations from https://www.statmechsims.com/models/metropolis to
    friendly format.

    params:
        - json_path (str): Path to JSON of saved simulation.
        - data_dir (str): Best to be the name of your experiment. Experiment directory where all images will be saved.
        - n_bins (int): Number of classes for data to be binned in. Default is None.
        - custom_intervals (np.ndarray): Set custom intervals for inequal boundaries, if desired. MUST be of shape (n_bins, 2) where each row is the
        upper and lower band of each interval. Default is None.
    '''
    def __init__(self, json_path:str, data_dir:str, n_bins:int=None, custom_intervals:np.ndarray=None) -> None:
        # Do standard initalization:
        self.json_path = json_path
        # Instantiate parser
        self.parser = jsonParser(self.json_path) 
        
        # Add n_bins and data_dir
        self.n_bins = n_bins
        self.data_dir = data_dir

        # Instantiate binner for continuity:
        self.binner = nBin(
            min_ = self.parser.temp_min,
            max_ = self.parser.temp_max,
            n_bins = self.n_bins
        )
        # This allows me to add custom_intervals to binner.intervals
        # It might change later
        self.binner.intervals = custom_intervals
        
        # If both are instantiated:
        if self.binner.intervals is not None and self.n_bins != None:
            # Ensure custom intervals line up with bins
            assert self.binner.intervals.shape[0] == n_bins, 'Number of specified bins must be equal to the number of rows in your custom intervals!'
        
        # If intervals are instantiated and n_bins are not instantiated:
        elif self.binner.intervals is not None and self.n_bins == None:
            # Just set number of bins to the number of rows in your intervals.
            self.n_bins = self.binner.intervals.shape[0]
        
        # If intervals are not instantiated and number of bins are instantiated:
        elif self.binner.intervals is None and self.n_bins != None:
            # Calculate equidistant range for each bin.
            self.bin_len = (self.parser.temp_max - self.parser.temp_min) / self.n_bins
            
            # Instantiate binner properly:
            self.binner = nBin(
                min_ = self.parser.temp_min,
                max_ = self.parser.temp_max,
                n_bins = self.n_bins,
                bin_len = self.bin_len
            )

            # Create interval bounds. Will be saved as self.binner.intervals of shape (n_bins, 2)
            self.binner._calculate_intervals()
        
        # Otherwise, neither are instantiated. You're gonna need at least one.
        else:
            assert (custom_intervals is not None and n_bins != None), 'Instantiate at least one parameter! Either n_bins or custom_intervals.'
    
    def _make_bins(self, data_dir_subdirectory:str):
        '''
        Make bin0...bin{n-1} directories.
        
        params:
            - data_dir_subdirectory (str): Path to subdirectory to create bins within.
        '''
        # Create new bins:
        for bin in range(self.n_bins):
            os.mkdir(os.path.join(data_dir_subdirectory, f'bin{bin}'))

    def _check_data_dir_exists(self, dir_name:str):
        '''
        Check if data directory exists. If it does, remove the directory tree and create a new one for bins to be inside. If it does not exist,
        create it and bins.

        params:
            - dir_name (str): Name of directory to see if it exists.
        '''
        # Check if data directory within experiment exists and remove it entirely.
        data_dir_subdirectory = os.path.join(self.data_dir, dir_name)
        if os.path.isdir(data_dir_subdirectory): 
            shutil.rmtree(data_dir_subdirectory)
        os.mkdir(data_dir_subdirectory)
    
    def _colormap(self, array_:np.ndarray):
        '''
        Convert array to image data
        
        params:
            - array_ (np.ndarray): 100x100 image array.
        '''
        cmap = {
            1: (255,255,255),
            -1: (0,0,0)
        }
        data = [cmap[spin] for spin in array_.flatten()]
        img = Image.new('RGB', (100,100), 'white')
        img.putdata(data)
        return img
    
    def _save_dataframe(self, image_col:list, labels:list, save_path:str):
        '''
        Quick function to save dataframes.
        
        params:
            - image_col (list): Data for image column.
            - labels (list): Data for labels.
            - save_path (str): Path to save the dataframe.
        '''
        pd.DataFrame({
            'image': image_col,
            'label': labels
        }).to_csv(save_path, index=None)

    def process(self):
        '''Process function to be used by end-user.'''
        # Check if data directory exists to empty out. If it does not exist, create it.
        self._check_data_dir_exists(dir_name='data')
        # Add bin0 ... bin(n-1):
        self._make_bins(os.path.join(self.data_dir, 'data'))
        # Iterate through data and match:
        # Will add a new key called bin_number when specified bin is created.
        # After this loop, a csv will be created:
        save_paths = []
        labels_ = []
        for data_info in self.parser.data:
            # Parse out temp
            temp = data_info['temp']
            # Save data array as image to data/bin_number:
            array_ = np.array([float(x)for x in data_info['spinArray'].split(',')]).reshape(100,100)
            # Some handling is done to convert to color mapping:
            img = self._colormap(array_)
            # Loop to through bins to find bin number:
            for bin_number in range(self.n_bins):
                if self.binner.intervals[bin_number, 0] <= temp <= self.binner.intervals[bin_number, 1]:
                    data_info['bin_number'] = bin_number
            # Construct save path
            save_path = os.path.join(self.data_dir, 'data', f'bin{data_info["bin_number"]}', f'{data_info["timestamp"]}.jpg')
            # Save array as JPEG
            img.save(save_path)
            # Add save paths and bin number to list for dataframe:
            save_paths.append(save_path)
            labels_.append(data_info["bin_number"])
        # Construct/save dataframe
        self._check_data_dir_exists('csvs')
        csv_path = os.path.join(self.data_dir, 'csvs', 'data.csv')
        self._save_dataframe(save_paths, labels_, csv_path)



'''Utils for model configuration. Most of these came from https://huggingface.co/blog/fine-tune-vit '''
import torch
import numpy as np
import pandas as pd
import os
from sklearn.model_selection import StratifiedShuffleSplit
from transformers import ViTForImageClassification, ViTFeatureExtractor
from datasets import load_metric, Dataset, Image

## DATASET RELATED UTILS

def check_for_dir(data_dir:str, dir_name:str):
    '''
    Check if directory within data_dir exists. If it doesn't create it and return the full path.
    '''
    dir_path = os.path.join(data_dir, dir_name)
    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)
    return dir_path

# The following functions will be used for stratified sampling:

def load_data_csv(data_dir:str) -> pd.DataFrame:
    return pd.read_csv(os.path.join(data_dir, 'csvs', 'data.csv'))

def stratified_shuffle(data_dir:str, num_labels:int, test_size:float) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    '''
    Run stratified shuffling method on binned data. Will return the train/test sets.
    '''
    # Load csv:
    df = load_data_csv(data_dir)
    # Instantiate shuffle split:
    sss = StratifiedShuffleSplit(num_labels, test_size=test_size)
    # Get X and y
    X, y = df['image'].to_numpy(), df['label'].to_numpy()
    # Stratified split and construct train/test sets
    for train_index, test_index in sss.split(X, y):
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]
    # Return the splits:
    return X_train, y_train, X_test, y_test


def load_dataset_from_csv(path_to_csv:str, apply_transform:bool=True) -> Dataset:
    '''
    Helper function to load validation set which has been saved as CSV file during training.
    
    params:
        - path_to_csv (str): Path to validation csv.
        - apply_transform (bool): Apply default transformation if True. Default set to True.
    
    returns:
        - dataset (Dataset): HuggingFace dataset.
    '''
    df = pd.read_csv(path_to_csv)
    # Load to HF Dataset
    dataset = Dataset.from_pandas(df).cast_column('image', Image())
    if apply_transform:
        dataset = dataset.with_transform(transform)
    return dataset

def transform(example_batch):
    ''' 
    "Perform" feature extraction on data. FE will actually run when the sample is indexed/called. In our case, it will likely be during
    training.
    '''
    feature_extractor = ViTFeatureExtractor(do_resize=True)
    # Converts the PNG file to processed array. This is the notation provided by HF in the link.
    inputs = feature_extractor([x for x in example_batch['image']], return_tensors='pt')
    # If there is an error, check the key for your label.
    inputs['label'] = example_batch['label']
    return inputs

def collate_fn(batch):
    '''Create collated data'''
    return {
        'pixel_values': torch.stack([x['pixel_values'] for x in batch]),
        'labels': torch.tensor([x['label'] for x in batch])
    }

## MODEL/TRAINING RELATED UTILS

def compute_metrics(p):
    '''Compute accuracy by geting MAX probability. Change to np.argmin if minimum prediction is needed.'''
    metric = load_metric("accuracy")
    return metric.compute(predictions=np.argmax(p.predictions, axis=1), references=p.label_ids)


def load_pretrained_vit_model(num_labels:int, pretrained_path:str='google/vit-base-patch32-224-in21k') -> ViTForImageClassification:
    '''
    Load pretrained ViT for training. Only use this if you want to reproduce results or use a different pretrained model. 
    You can find the list of vision transformers here: https://huggingface.co/models?sort=downloads&search=google%2Fvit-
    
    params:
        - num_labels (int): Number of prediction labels.
        - pretrained_path (str): Either HF URL to ViT for image classification OR path to local pretrained model.
    
    returns:
        - model (ViTForImageClassification): Pretrained ViT model for image classification tasks.
    '''
    model = ViTForImageClassification.from_pretrained(pretrained_path, num_labels = num_labels)
    return model